@@ -22,6 +22,7 @@ from .core import (
     render_fixed,
     use_context,
     use_effect,
+    use_exception,
     use_memo,
     use_reducer,
     use_ref,
@@ -40,6 +41,7 @@ __all__ = [
     "get_widget",
     "use_context",
     "create_context",
+    "use_exception",
     "use_memo",
     "use_ref",
     "use_state",
