@@ -1,5 +1,7 @@
+import time
 import traceback
 import unittest.mock
+from concurrent.futures import ThreadPoolExecutor
 from typing import Callable, List, Optional, Tuple, TypeVar, cast
 
 import ipywidgets
@@ -59,7 +61,7 @@ def ButtonNumber2(value):
 
 
 @react.component
-def Container(children=[]):
+def ContainerFunction(children=[]):
     return w.HBox(children=children)
 
 
@@ -75,12 +77,16 @@ def cleanup_guard():
         # raise RuntimeError(f"{leftover_widgets}")
 
 
-# @pytest.fixture(params=["ButtonComponentWidget", "ButtonComponentFunction"])
-@pytest.fixture(params=["ButtonComponentWidget"])
+@pytest.fixture(params=["ButtonComponentWidget", "ButtonComponentFunction"])
 def ButtonComponent(request):
     return dict(ButtonComponentWidget=w.Button, ButtonComponentFunction=ButtonComponentFunction)[request.param]
 
 
+@pytest.fixture(params=["ContainerWidget", "ContainerFunction"])
+def Container(request):
+    return dict(ContainerWidget=w.HBox, ContainerFunction=ContainerFunction)[request.param]
+
+
 def test_internals():
     @react.component
     def Child():
@@ -98,6 +104,7 @@ def test_internals():
     #     - '/':
     widget, rc = react.render_fixed(app, handle_error=False)
     assert rc.context_root.root_element == app
+    assert rc.context_root.root_element_next is None
     assert list(rc.context_root.children_next) == []
     assert list(rc.context_root.children) == ["App/"]
 
@@ -123,6 +130,173 @@ def test_create_element():
     rc.close()
 
 
+def test_no_on_value_on_write():
+    on_value = unittest.mock.Mock()
+
+    def set_value(x: int):
+        pass
+
+    @react.component
+    def Test():
+        nonlocal set_value
+        value, set_value = react.use_state(0)  # type: ignore
+        return w.IntSlider(value=value, on_value=on_value)
+
+    box, rc = react.render(Test(), handle_error=False)
+    assert len(rc._find(widgets.IntSlider)) == 1
+    set_value(1)
+    on_value.assert_not_called()
+    rc.close()
+
+
+def test_render_element_twice(ButtonComponent, Container):
+    el = ButtonComponent(description="Hi")
+
+    @react.component
+    def Test(el):
+        return Container(children=[el, el])
+
+    box, rc = react.render(Test(el), handle_error=False)
+    assert len(rc._find(widgets.Button)) == 2
+    rc.force_update()
+    assert len(rc._find(widgets.Button)) == 2
+    rc.force_update()
+    assert len(rc._find(widgets.Button)) == 2
+    rc.close()
+
+
+def test_remove_element_twice(ButtonComponent, Container):
+    el1 = ButtonComponent(description="Hi1")
+    el2 = w.FloatLogSlider(description="Hi2")
+
+    def set_first(x: bool):
+        pass
+
+    @react.component
+    def Test(el1, el2):
+        nonlocal set_first
+        first, set_first = react.use_state(True)  # type: ignore
+        if first:
+            el = el1
+        else:
+            el = el2
+        # return Container(children=[Container(children=[el])], Container(children=[el]))
+        with Container(children=[el]) as main:
+            with Container(children=[el]):
+                pass
+            with Container(children=[el]):
+                pass
+        return main
+
+    box, rc = react.render(Test(el1, el2), handle_error=False)
+    assert len(rc._find(widgets.Button)) == 3
+    rc.force_update()
+    assert len(rc._find(widgets.Button)) == 3
+    rc.force_update()
+    assert len(rc._find(widgets.Button)) == 3
+    assert rc._find(widgets.Button)[0].widget.description == "Hi1"
+    set_first(False)
+    assert len(rc._find(widgets.FloatLogSlider)) == 3
+    assert rc._find(widgets.FloatLogSlider)[0].widget.description == "Hi2"
+    rc.force_update()
+    rc.close()
+
+
+def test_remove_element_repeated(ButtonComponent, Container):
+    def set_other_case(x: bool):
+        pass
+
+    @react.component
+    def Test():
+        nonlocal set_other_case
+        el = ButtonComponent(description="Hi")
+        other_case, set_other_case = react.use_state(True)  # type: ignore
+        with Container(children=[el]) as main:
+            with Container(children=[el]):
+                pass
+            if other_case:
+                # this will trigger the removal of el twice
+                w.Label(value="bump up the children")
+            with Container(children=[el]):
+                pass
+            with Container(children=[el]):
+                pass
+        return main
+
+    box, rc = react.render(Test(), handle_error=False)
+    assert len(rc._find(widgets.Button)) == 4
+    buttons = rc._find(widgets.Button).widgets
+    assert [b is not buttons[0] for b in buttons]
+    rc.force_update()
+    buttons = rc._find(widgets.Button).widgets
+    assert [b is not buttons[0] for b in buttons]
+    assert len(rc._find(widgets.Button)) == 4
+    buttons = rc._find(widgets.Button).widgets
+    assert [b is not buttons[0] for b in buttons]
+    rc.force_update()
+    buttons = rc._find(widgets.Button).widgets
+    assert [b is not buttons[0] for b in buttons]
+    assert len(rc._find(widgets.Button)) == 4
+    assert rc._find(widgets.Button)[0].widget.description == "Hi"
+    set_other_case(False)
+    rc.force_update()
+    rc.close()
+
+
+def test_replace_parent(ButtonComponent):
+    def set_other_case(x: bool):
+        pass
+
+    @react.component
+    def Test():
+        nonlocal set_other_case
+        other_case, set_other_case = react.use_state(True)  # type: ignore
+        el = ButtonComponent(description="Hi")
+        if other_case:
+            Container = w.VBox
+        else:
+            Container = w.HBox
+        # the root stays the same to have the keys the same
+        with w.VBox() as main:
+            # but the intermediate parent changes
+            with Container(children=[el]):
+                pass
+        return main
+
+    box, rc = react.render(Test(), handle_error=False)
+    assert len(rc._find(widgets.Button)) == 1
+    rc.force_update()
+    set_other_case(False)
+    rc.force_update()
+    rc.close()
+
+
+@pytest.mark.parametrize("shared", [False, True])
+def test_render_count_element(Container, ButtonComponent, shared):
+    button = ButtonComponent()
+
+    @react.component
+    def DoNothing(children=[]):
+        return ButtonComponent(description="ignore the arguments")
+
+    @react.component
+    def Test():
+        nonlocal button
+        button = ButtonComponent(description="Button")
+        if shared:
+            button = button.shared()
+        with Container() as main:
+            w.HBox(children=[button])
+            w.VBox(children=[button])
+            Container(children=[button])
+            DoNothing(children=[button])
+        return main
+
+    box, rc = react.render(Test(), handle_error=False)
+    assert button._render_count == (1 if shared else 3)
+    rc.close()
+
+
 def test_monkey_patch():
     button = widgets.Button.element(description="Hi")
     hbox, rc = react.render(button)
@@ -159,6 +333,88 @@ def test_component_function():
     rc.close()
 
 
+def test_render_replace():
+    hbox, rc = react.render(w.Button())
+    rc.render(w.VBox(), rc.container)
+    rc.close()
+
+
+def test_render_replace_component_with_element():
+
+    set_value = lambda x: None  # noqa
+
+    def Container():
+        button = w.Button(description="Button in container")
+        return w.VBox(children=[button])
+
+    @react.component
+    def Test():
+        nonlocal set_value
+        value, set_value = react.use_state(0)
+        if value == 1:
+            return Container()
+        else:
+            return w.Button(description="Button in root")
+
+    hbox, rc = react.render(Test(), handle_error=False)
+    rc._find(widgets.Button).widget.description == "Button in root"
+    set_value(1)
+    rc._find(widgets.Button).widget.description == "Button in container"
+    set_value(0)
+    rc._find(widgets.Button).widget.description == "Button in root"
+    rc.close()
+
+
+def test_render_many_flip_flop_component():
+    set_action = None
+
+    effect_mock = unittest.mock.Mock()
+    cleanup = unittest.mock.Mock()
+
+    @react.component
+    def Component(description):
+        description, set_description = react.use_state(description)
+
+        def effect():
+            effect_mock()
+            return cleanup
+
+        # also test use effect to be sure!
+        react.use_effect(effect)
+        return w.Button(description=description)
+
+    @react.component
+    def Test():
+        nonlocal set_action
+        action, set_action = react.use_state(0)
+        if action == 0:
+            return w.Text(value="initial")
+        elif action == 1:
+            set_action(2)
+            # render component
+            return Component("first")
+        elif action == 2:
+            # render float slider next
+            set_action(3)
+            return w.FloatSlider()
+        elif action == 3:
+            # set_action(4)
+            # and return to the component again
+            return Component("second")
+        else:
+            return w.FloatLogSlider()
+
+    box, rc = react.render(Test())
+    assert isinstance(box.children[0], widgets.Text)
+    assert set_action is not None
+    set_action(1)
+
+    assert isinstance(box.children[0], widgets.Button)
+    assert box.children[0].description == "second"
+    assert react.core.local.last_rc is not None
+    react.core.local.last_rc.close()
+
+
 def test_state_simple():
     clear()
 
@@ -344,19 +600,18 @@ def test_children():
 
 
 def test_display():
-    slider = widgets.IntSlider(value=2, description="How many buttons?")
-
     @react.component
-    def buttons(slider):
-        buttons = react.use_state_widget(slider, "value")
-        return w.Button(description=f"Button {buttons}")
+    def Button(value):
+        return w.Button(description=f"Value {value}")
 
-    react.display(buttons(slider))
-    assert react.core._last_rc is not None
-    react.core._last_rc.close()
-    slider.style.close()
-    slider.layout.close()
-    slider.close()
+    react.display(Button(2))
+    assert react.core.local.last_rc is not None
+    vbox: widgets.VBox = react.core.local.last_rc.container
+    assert vbox._view_count == 0
+    vbox._view_count = 1  # act as if it is displayed
+    vbox._view_count = 0  # and removed again
+    # so we don't have to do this manually:
+    # react.core.local.last_rc.close()
 
 
 def test_box():
@@ -396,17 +651,17 @@ def test_box():
     slider.close()
 
 
-def test_shared_instance(ButtonComponent):
+def test_shared_instance_via_widget_element(ButtonComponent, Container):
     checkbox = widgets.Checkbox(value=True, description="Share button")
 
     @react.component
     def Buttons(checkbox):
         share = react.use_state_widget(checkbox, "value", "share")
         if share:
-            button_shared = ButtonComponent(description="Button shared", tooltip="shared")
-            return w.VBox(children=[button_shared, button_shared])
+            button_shared = ButtonComponent(description="Button shared", tooltip="shared").shared()
+            return Container(children=[button_shared, button_shared])
         else:
-            return w.VBox(children=[ButtonComponent(description=f"Button {i}") for i in range(2)])
+            return Container(children=[ButtonComponent(description=f"Button {i}") for i in range(2)])
 
     hbox, rc = react.render(Buttons(checkbox))
     vbox = hbox.children[0]
@@ -427,7 +682,8 @@ def test_shared_instance(ButtonComponent):
     checkbox.close()
 
 
-def test_shared_instance_via_component(ButtonComponent):
+@pytest.mark.parametrize("shared", [False, True])
+def test_shared_instance_via_component(ButtonComponent, shared, Container):
     @react.component
     def Child(button):
         return button
@@ -435,12 +691,17 @@ def test_shared_instance_via_component(ButtonComponent):
     @react.component
     def Buttons():
         button = ButtonComponent(description="Button shared")
-        return w.VBox(children=[Child(button), Child(button)])
+        if shared:
+            button = button.shared()
+        return Container(children=[Child(button), Child(button)])
 
     vbox, rc = react.render_fixed(Buttons())
     assert vbox.children[0].description == "Button shared"
     assert vbox.children[1].description == "Button shared"
-    assert vbox.children[0] is vbox.children[1]
+    if shared:
+        assert vbox.children[0] is vbox.children[1]
+    else:
+        assert vbox.children[0] is not vbox.children[1]
     rc.close()
 
 
@@ -454,8 +715,8 @@ def test_bqplot():
     def Plot(exponent, x, y):
         exponent_value = react.use_state_widget(exponent, "value")
         x = x**exponent_value
-        x_scale = bqplot.LinearScale(allow_padding=False)
-        y_scale = bqplot.LinearScale(allow_padding=False)
+        x_scale = bqplot.LinearScale(allow_padding=False).shared()
+        y_scale = bqplot.LinearScale(allow_padding=False).shared()
         lines = bqplot.Lines(x=x, y=y, scales={"x": x_scale, "y": y_scale}, stroke_width=3, colors=["red"], display_legend=True, labels=["Line chart"])
         x_axis = bqplot.Axis(scale=x_scale)
         y_axis = bqplot.Axis(scale=y_scale)
@@ -512,7 +773,7 @@ def test_use_effect():
         button_el = w.Button(description=f"Clicked {clicks} times")
         return button_el
 
-    hbox, rc = react.render(Button2())
+    hbox, rc = react.render(Button2(), handle_error=False)
     assert count() == 2 + 3  # label + button
     button = hbox.children[0]
     assert button.description == "Clicked 0 times"
@@ -653,8 +914,8 @@ def test_key_widget():
         return main
 
     box = react.make(Buttons(), handle_error=False)
-    assert react.core._last_rc
-    rc = react.core._last_rc
+    assert react.core.local.last_rc
+    rc = react.core.local.last_rc
     assert set_reverse is not None
     button1, slider1 = box.children[0].children
     assert isinstance(button1, widgets.Button)
@@ -666,7 +927,7 @@ def test_key_widget():
     assert isinstance(slider2, widgets.IntSlider)
     assert button1 is button2
     assert slider1 is slider2
-    assert react.core._last_rc
+    assert react.core.local.last_rc
     rc.close()
 
 
@@ -678,8 +939,8 @@ def test_key_root():
     box = react.make(Buttons(), handle_error=False)
     button = box.children[0]
     assert isinstance(button, widgets.Button)
-    assert react.core._last_rc
-    react.core._last_rc.close()
+    assert react.core.local.last_rc
+    react.core.local.last_rc.close()
 
 
 def test_key_component_function():
@@ -701,8 +962,8 @@ def test_key_component_function():
         return w.HBox(children=[slider, checkbox, buttons_box])
 
     box = react.make(Buttons(), handle_error=False)
-    assert react.core._last_rc
-    rc = react.core._last_rc
+    assert react.core.local.last_rc
+    rc = react.core.local.last_rc
     slider, checkbox, buttons = box.children[0].children
     assert buttons.children[0].description == "0: Clicked 0 times"
     assert buttons.children[1].description == "1: Clicked 0 times"
@@ -777,8 +1038,8 @@ def test_interactive():
     hbox = box.children[0]
     button0 = hbox.children[0]
     assert button0.description == "Button 0"
-    assert react.core._last_rc is not None
-    react.core._last_rc.close()
+    assert react.core.local.last_rc is not None
+    react.core.local.last_rc.close()
     for widget in control.children:
         if hasattr(widget, "layout"):
             widget.layout.close()
@@ -839,6 +1100,7 @@ def test_context():
     @react.component
     def App():
         clicks, dispatch = react.use_reducer(click_reducer, 0)
+        assert react.use_context(store_context) is None
         store_context.provide((clicks, dispatch))
         with w.HBox() as main:
             Child1()
@@ -862,49 +1124,132 @@ def test_context():
     rc.close()
 
 
+# @pytest.mark.parametrize("shared", [False, True])
+@pytest.mark.parametrize("shared", [False])
+def test_context_with_precreated_element(shared):
+    value_context = react.create_context(1)
+
+    @react.component
+    def ValueProvider(value: int, children=[]):
+        value_context.provide(value)
+        return w.VBox(children=children)
+
+    @react.component
+    def Button():
+        value = react.use_context(value_context)
+        return w.Button(description=f"Value: {value}")
+
+    @react.component
+    def Test():
+        button = Button()
+        if shared:
+            button = button.shared()
+        v = ValueProvider(42, children=[button])
+        return v
+
+    box, rc = react.render(Test())
+    button = rc._find(widgets.Button).widget
+    if shared:
+        assert button.description == "Value: 1"
+    else:
+        assert button.description == "Value: 42"
+    rc.close()
+
+
 def test_memo():
-    calls_ab = 0
-    calls_ac = 0
+    __calls_ab = 0
+    __calls_ac = 0
 
     @react.component
     def TestMemo(a, b, c):
         def expensive_ab(i, j):
-            nonlocal calls_ab
-            calls_ab += 1
+            nonlocal __calls_ab
+            __calls_ab += 1
             return i + j
 
-        @react.use_memo
         def expensive_ac(i, j):
-            nonlocal calls_ac
-            calls_ac += 1
+            nonlocal __calls_ac
+            __calls_ac += 1
             return i + j * 2
 
-        x = react.use_memo(expensive_ab, args=[a], kwargs={"j": b})
-        y = expensive_ac(a, c)
+        x = react.use_memo(lambda: expensive_ab(a, b), [a, b])
+        y = react.use_memo(lambda: expensive_ac(a, c), [a, c])
         return w.Label(value=f"{x} - {y}")
 
     label, rc = react.render_fixed(TestMemo(a=1, b=2, c=3))
-    assert calls_ab == 1
-    assert calls_ac == 1
+    assert __calls_ab == 1
+    assert __calls_ac == 1
     assert label.value == "3 - 7"
 
     rc.render(TestMemo(a=1, b=20, c=3))
-    assert calls_ab == 2
-    assert calls_ac == 1
+    assert __calls_ab == 2
+    assert __calls_ac == 1
     assert label.value == "21 - 7"
 
     rc.render(TestMemo(a=1, b=20, c=30))
-    assert calls_ab == 2
-    assert calls_ac == 2
+    assert __calls_ab == 2
+    assert __calls_ac == 2
     assert label.value == "21 - 61"
 
     rc.render(TestMemo(a=10, b=20, c=30))
-    assert calls_ab == 3
-    assert calls_ac == 3
+    assert __calls_ab == 3
+    assert __calls_ac == 3
     assert label.value == "30 - 70"
     rc.close()
 
 
+def test_memo_auto_closure():
+    __calls = 0
+
+    @react.component
+    def Test(value):
+        def square():
+            nonlocal __calls
+            __calls += 1
+            return value**2
+
+        y = react.use_memo(square)
+        return w.Label(value=f"{value} - {y}")
+
+    box, rc = react.render(Test(value=1))
+    assert __calls == 1
+    rc.render(Test(1))
+    assert __calls == 1
+    rc.render(Test(2))
+    assert __calls == 2
+    rc.render(Test(2))
+    assert __calls == 2
+    rc.render(Test(1))
+    assert __calls == 3
+    rc.close()
+
+
+def test_memo_like_react():
+    __calls = 0
+
+    @react.component
+    def Test(value):
+        def square():
+            nonlocal __calls
+            __calls += 1
+            return value**2
+
+        y = react.use_memo(square, [value])
+        return w.Label(value=f"{value} - {y}")
+
+    box, rc = react.render(Test(value=1))
+    assert __calls == 1
+    rc.render(Test(1))
+    assert __calls == 1
+    rc.render(Test(2))
+    assert __calls == 2
+    rc.render(Test(2))
+    assert __calls == 2
+    rc.render(Test(1))
+    assert __calls == 3
+    rc.close()
+
+
 def test_container_context_simple():
     @react.component
     def ContainerContext():
@@ -925,8 +1270,8 @@ def test_container_context_bqplot():
         x = np.arange(4)
         y = x**1.2
         with w.HBox() as box:
-            x_scale = bqplot.LinearScale(allow_padding=False)
-            y_scale = bqplot.LinearScale(allow_padding=False)
+            x_scale = bqplot.LinearScale(allow_padding=False).shared()
+            y_scale = bqplot.LinearScale(allow_padding=False).shared()
             lines = bqplot.Lines(x=x, y=y, scales={"x": x_scale, "y": y_scale}, stroke_width=3, colors=["red"], display_legend=True, labels=["Line chart"])
             x_axis = bqplot.Axis(scale=x_scale)
             y_axis = bqplot.Axis(scale=y_scale)
@@ -1223,9 +1568,32 @@ def test_render_many_consolidate_once():
     rc.close()
 
 
-def test_recover_exception(capsys):
+def test_exception_recover_in_render():
     set_fail = None
 
+    # by default react should render the exception
+    @react.component
+    def Test():
+        nonlocal set_fail
+        fail, set_fail = react.use_state(False)
+        if fail:
+            raise Exception("fail")
+        return w.IntSlider()
+
+    box, rc = react.render(Test(), handle_error=True)
+    assert set_fail is not None
+    assert rc._find(ipywidgets.IntSlider)
+    set_fail(True)
+    assert not rc._find(ipywidgets.IntSlider)
+    assert not rc._is_rendering
+    assert "Traceback" in rc._find(ipywidgets.HTML).widget.value
+    rc.close()
+
+
+def test_exception_handler_in_render():
+    set_fail = None
+    clear = None
+
     @react.component
     def Test():
         nonlocal set_fail
@@ -1234,14 +1602,115 @@ def test_recover_exception(capsys):
             raise Exception("fail")
         return w.IntSlider()
 
-    # with pytest.raises(Exception):
-    slider, rc = react.render_fixed(Test())
+    @react.component
+    def Handler():
+        nonlocal clear
+        exception, clear = react.use_exception()
+        with w.VBox() as main:
+            if exception:
+                w.Label(value=str(exception))
+            else:
+                Test()
+        return main
+
+    box, rc = react.render(Handler())
     assert set_fail is not None
-    with pytest.raises(Exception, match="fail"):
-        set_fail(True)
+    assert clear is not None
+    assert rc._find(ipywidgets.IntSlider)
+
+    set_fail(True)
+    assert "fail" in rc._find(ipywidgets.Label).widget.value
+    assert not rc._is_rendering
+
+    # we keep the UI the same
     set_fail(False)
+    assert "fail" in rc._find(ipywidgets.Label).widget.value
+
+    # we have to explicitly clear the exception to try a rerender
+    clear()
+    assert rc._find(ipywidgets.IntSlider)
+    rc.close()
+
+
+def test_exception_handler_exception():
+    # when the exception handling raises an exception, rerender again and bubble up
+    set_fail = None
+    clear = None
+
+    @react.component
+    def Test():
+        nonlocal set_fail
+        fail, set_fail = react.use_state(False)
+        if fail:
+            raise Exception("fail")
+        return w.IntSlider()
+
+    @react.component
+    def Handler():
+        nonlocal clear
+        exception, clear = react.use_exception()
+        if exception:
+            # this handler should not be handled by the current component
+            raise Exception("in error handler")
+        else:
+            return Test()
+
+    box, rc = react.render(Handler())
+    assert set_fail is not None
+    assert clear is not None
+    assert rc._find(ipywidgets.IntSlider)
+    set_fail(True)
+    # because the Handler failed to catch the error, the core library does it
+    # using a HTML widget
+    assert "Traceback" in rc._find(ipywidgets.HTML).widget.value
+    assert not rc._is_rendering
+
+    rc.close()
+
+
+def test_recover_exception_in_reconcilliate():
+    set_fail = None
+
+    @react.component
+    def Test():
+        nonlocal set_fail
+        fail, set_fail = react.use_state(False)
+
+        def some_effect():
+            if fail:
+                raise Exception("fail")
+
+        react.use_effect(some_effect, [fail])
+        return w.IntSlider()
+
+    box, rc = react.render(Test())
+    assert set_fail is not None
+    set_fail(True)
+    assert "Traceback" in rc._find(ipywidgets.HTML).widget.value
+    assert not rc._is_rendering
+    rc.close()
+
+
+def test_recover_exception_in_widget_update():
+    set_fail = None
+
+    @react.component
+    def Test():
+        nonlocal set_fail
+        fail, set_fail = react.use_state(False)
+        value = 1
+        if fail:
+            value = "fail"  # type: ignore
+        return w.IntSlider(value=value)
+
+    # with pytest.raises(Exception):
+    box, rc = react.render(Test())
+    assert rc._find(ipywidgets.IntSlider)
+    assert set_fail is not None
+    set_fail(True)
+    assert not rc._is_rendering
+    assert not rc._find(ipywidgets.IntSlider)
     rc.close()
-    assert isinstance(slider, ipywidgets.IntSlider)
 
 
 def test_state_get():
@@ -1475,7 +1944,7 @@ def test_vue_orphan_not_close():
 
 
 @pytest.mark.parametrize("in_container", [False, True])
-def test_switch_component(in_container):
+def test_switch_component(in_container, Container):
     @react.component
     def Child1():
         with Container() as main:
@@ -1540,7 +2009,7 @@ def test_switch_component(in_container):
     rc.close()
 
 
-def test_switch_simple():
+def test_switch_simple(Container):
     set_value = None
 
     @react.component
@@ -1563,7 +2032,7 @@ def test_switch_simple():
     rc.close()
 
 
-def test_switch_widget_and_component():
+def test_switch_widget_and_component(Container):
     set_value = None
 
     effect = unittest.mock.Mock()
@@ -1612,7 +2081,7 @@ def test_switch_widget_and_component():
     rc.close()
 
 
-def test_switch_component_key():
+def test_switch_component_key(Container):
     set_value = None
 
     effect = unittest.mock.Mock()
@@ -1685,5 +2154,58 @@ def test_render_twice_different_element():
     set_action(1)
 
     assert isinstance(box.children[0], widgets.Text)
-    assert react.core._last_rc is not None
-    react.core._last_rc.close()
+    assert react.core.local.last_rc is not None
+    react.core.local.last_rc.close()
+
+
+def test_multithreaded_support():
+    N = 1000
+
+    @react.component
+    def Test(i):
+        value, set_value = react.use_state(10)
+        # trigger a few calls to set_value which uses the thread local
+        # render context
+        if value == 10:
+            # a bit of sleep so we get real concurrency
+            time.sleep(0.001)
+            set_value(5)
+        if value == 5:
+            time.sleep(0.001)
+            set_value(1)
+        return w.Button(description=str(i))
+
+    def worker(i):
+        button, rc = react.render_fixed(Test(i))
+
+        assert button.description == str(i)
+        return button, rc
+
+    pool = ThreadPoolExecutor(max_workers=N)
+    results = pool.map(worker, range(100))
+
+    for _button, rc in results:
+        rc = cast(react.core._RenderContext, rc)
+        rc.close()
+
+
+def test_nested_render():
+    @react.component
+    def Test():
+        box, rc = react.render(w.Button())
+        button = box.children[0]
+
+        def only_cleanup():
+            def cleanup():
+                rc.close()
+
+            return cleanup
+
+        react.use_effect(only_cleanup, [])
+        # note that here we pass in a widget, not an element
+        # this is supported, but not recommended in general
+        return w.VBox(children=[button])
+
+    box, rc = react.render(Test())
+    assert len(rc._find(widgets.Button)) == 1
+    rc.close()
